def write_to_dat(filename, data):
    if not filename.endswith('.dat'):
        raise ValueError(f'Filename must end with .dat extension but was {filename}')
    f = open(filename, 'w')
    for i in data:
        f.write(f"{i:.4f}\n")
    f.close()



import os

import vtk


def vtk_polydata_writer(filename, data, store_binary=False, store_xml=False):
    """Writes a vtk polydata grid to specified filename this has to include the full path"""
    if not (filename.endswith('.vtk') or filename.endswith(".vtp")):
        raise ValueError(f'Filename must end with .vtk or .vtp but was {filename}')
    if store_xml:
        writer = vtk.vtkXMLPolyDataWriter()
    else:
        writer = vtk.vtkPolyDataWriter()
    writer.SetFileName(filename)
    writer.SetInputData(data)
    if store_binary:
        writer.SetFileTypeToBinary()
    writer.Write()


def vtk_unstructured_grid_writer(filename, data, store_binary=False):
    if not (filename.endswith('.vtk') or filename.endswith(".vtu")):
        raise ValueError(f'Filename must end with .vtk or .vtu but was {filename}')
    writer = vtk.vtkUnstructuredGridWriter()
    writer.SetFileName(filename)
    writer.SetInputData(data)
    if store_binary:
        writer.SetFileTypeToBinary()
    writer.Write()


def vtk_xml_unstructured_grid_writer(filename, data):
    if not (filename.endswith('.vtk') or filename.endswith(".vtu")):
        raise ValueError(f'Filename must end with .vtk or .vtu but was {filename}')
    writer = vtk.vtkXMLUnstructuredGridWriter()
    writer.SetFileName(filename)
    writer.SetInputData(data)
    writer.Write()


def vtk_obj_writer(filename, data):
    if not filename.endswith('.obj'):
        raise ValueError(f'Filename must end with .obj but was {filename}')
    writer = vtk.vtkOBJWriter()
    writer.SetFileName(filename)
    writer.SetInputData(data)
    writer.Write()


def write_to_vtx(filename, data, try_append=False):
    """
    Writes the data to a .vtx file
    :param filename: Filename to write the data to with .vtx extension
    :param data: The data to write. Can be a list/array or single values
    :param try_append: Set to true if you like to append to existing .vtx file
    :return:
    """
    if not filename.endswith('.vtx'):
        raise ValueError(f'Filename must end with .vtx but was {filename}')
    length_data = len(data) if hasattr(data, '__len__') else 1
    if try_append and os.path.exists(filename):
        f = open(filename, 'a')
    else:
        f = open(filename, 'w')
        f.write(f'{length_data}\n')
        f.write('extra\n')

    if hasattr(data, '__iter__') or hasattr(data, '__getitem__'):
        for i in data:
            f.write(f'{i}\n')
    else:
        f.write(f'{data}\n')
    f.close()



import vtk
from vtkmodules.util.numpy_support import vtk_to_numpy as orig_vtk_numpy
from vtkmodules.util.numpy_support import numpy_to_vtk as orig_numpy_vtk

def vtk_to_numpy(vtk_data):
    return orig_vtk_numpy(vtk_data)


def numpy_to_vtk(numpy_data, deep=0, array_type=None):
    return orig_numpy_vtk(numpy_data, deep=deep, array_type=array_type)


def convert_point_to_cell_data(model, point_array_names=None):
    """
    Converts point wise data to cell data
    """

    pt_cell = vtk.vtkPointDataToCellData()
    pt_cell.SetInputData(model)
    if point_array_names is not None:
        pt_cell.CategoricalDataOff()
        pt_cell.ProcessAllArraysOff()
        for array_names in point_array_names:
            pt_cell.AddPointDataArray(array_names)
    pt_cell.PassPointDataOn()

    pt_cell.Update()

    return pt_cell.GetOutput()



import pyvista as pv


def pick_point_with_preselection(mesh, object_str, current_point, point_scale=1):
    """
    Pick a point on a mesh interactively and select an orientation point
    :param mesh: The mesh where a point should be picked
    :param object_str: The name of the object which should be picked
    :param current_point: The current guess where the point is. Will be visualized.
    :param point_scale: Scale of the current point on the model
    :return:
    """
    p = pv.Plotter(notebook=False)
    if current_point is not None:
        point_cloud = pv.PolyData(current_point)
        p.add_mesh(point_cloud, color='w', point_size=30. * point_scale, render_points_as_spheres=True)
    p.add_mesh(mesh, color='r')
    p.enable_point_picking(mesh, use_picker=True)
    p.add_text(f'Select the {object_str} and close the window', position='lower_left')
    p.show()
    if p.picked_point is not None:
        current_point = p.picked_point
    elif current_point is not None:
        return current_point
    else:
        raise ValueError(f"Please select the {object_str}")
    return current_point


def pick_point(mesh, object_str):
    """
    Pick a point on a mesh interactively
    :param mesh: The mesh where a point should be picked
    :param object_str: The name of the object which should be picked
    :return:
    """
    return pick_point_with_preselection(mesh, object_str, None)



from src.vtk_openCARP_methods_ibt.carputils_commands.lat_command import LATCommand


def test_lat_command_initialization_consistency():
    # Create initial command with all attributes
    initial_cmd = LATCommand(
        threshold=0.9,
        id_name="ACTs",
        positional_index=0,
        all_flag=0,
        measurand=0,
        mode=0
    )

    # Convert to command string and back to class
    cmd_str = initial_cmd.to_cmd()
    converted_cmd = LATCommand.create_from_cmd_list(cmd_str)

    # Check if all attributes are preserved
    assert initial_cmd.threshold == converted_cmd.threshold
    assert initial_cmd.id_name == converted_cmd.id_name
    assert initial_cmd.positional_index == converted_cmd.positional_index
    assert initial_cmd.all == converted_cmd.all
    assert initial_cmd.measurand == converted_cmd.measurand
    assert initial_cmd.mode == converted_cmd.mode

    # Check if all attributes are identical
    assert vars(initial_cmd) == vars(converted_cmd)



from pathlib import Path

import numpy as np
import vtk
from scipy.spatial import KDTree
from vtkmodules.numpy_interface import dataset_adapter as dsa

from ..openCARP.exporting import write_to_pts
from ..vtk_methods.converters import vtk_to_numpy
from ..vtk_methods.exporting import vtk_unstructured_grid_writer, write_to_vtx
from ..vtk_methods.reader import vtx_reader, smart_reader


def point_array_mapper(mesh1, mesh2, idat):
    """

    :param mesh1:
    :param mesh2:
    :param idat: Set to "all" to mapp all arrays, else choose a specific array
    :return: Mapped mesh
    """
    pts1 = vtk_to_numpy(mesh1.GetPoints().GetData())
    pts2 = vtk_to_numpy(mesh2.GetPoints().GetData())

    meshNew = _point_array_mapper(idat, mesh1, mesh2, pts1, pts2)

    return meshNew.VTKObject


def _point_array_mapper(idat, mesh1, mesh2, pts1, pts2):
    """

    :param idat: Set to "all" to mapp all arrays, else choose a specific array
    :param mesh1:
    :param mesh2:
    :param pts1:
    :param pts2:
    :return:
    """
    tree = KDTree(pts1)
    dd, ii = tree.query(pts2, workers=-1)
    meshNew = dsa.WrapDataObject(mesh2)
    if idat == "all":
        for i in range(mesh1.GetPointData().GetNumberOfArrays()):
            data = vtk_to_numpy(
                mesh1.GetPointData().GetArray(mesh1.GetPointData().GetArrayName(i)))
            data2 = data[ii]
            data2 = np.where(np.isnan(data2), 10000, data2)

            meshNew.PointData.append(data2, mesh1.GetPointData().GetArrayName(i))
    else:
        data = vtk_to_numpy(mesh1.GetPointData().GetArray(idat))
        data2 = data[ii]
        meshNew.PointData.append(data2, idat)
    return meshNew


def cell_array_mapper(mesh1, mesh2, mesh2_name, idat):
    """

    :param mesh1: Vtk mesh object
    :param mesh2: Vtk mesh object
    :param mesh2_name:
    :param idat: Set to "all" to mapp all arrays, else choose a specific array
    :return: Returns mesh and also stores it to disk with subfix "_with_data"
    """
    filter_cell_centers = vtk.vtkCellCenters()
    filter_cell_centers.SetInputData(mesh1)
    filter_cell_centers.Update()
    centroids1 = filter_cell_centers.GetOutput().GetPoints()
    centroids1_array = vtk_to_numpy(centroids1.GetData())

    filter_cell_centers = vtk.vtkCellCenters()
    filter_cell_centers.SetInputData(mesh2)
    filter_cell_centers.Update()
    centroids2 = filter_cell_centers.GetOutput().GetPoints()
    pts2 = vtk_to_numpy(centroids2.GetData())

    meshNew = _cell_array_mapper(idat, mesh1, mesh2, centroids1_array, pts2)

    vtk_unstructured_grid_writer(f"{mesh2_name.split('.')[0]}_with_data.vtk", meshNew.VTKObject)

    return meshNew.VTKObject


def _cell_array_mapper(idat, mesh1, mesh2, array1, array2):
    """

    :param idat: Set to "all" to mapp all arrays, else choose a specific array
    :param mesh1:
    :param mesh2:
    :param array1:
    :param array2:
    :return:
    """
    tree = KDTree(array1)
    dd, ii = tree.query(array2, workers=-1)
    meshNew = dsa.WrapDataObject(mesh2)
    if idat == "all":
        for i in range(mesh1.GetCellData().GetNumberOfArrays()):
            data = vtk_to_numpy(
                mesh1.GetCellData().GetArray(mesh1.GetCellData().GetArrayName(i)))
            data2 = data[ii]
            meshNew.CellData.append(data2, mesh1.GetCellData().GetArrayName(i))
    else:
        data = vtk_to_numpy(mesh1.GetCellData().GetArray(idat))
        data2 = data[ii]
        meshNew.CellData.append(data2, idat)
    return meshNew


def mapp_ids_for_folder(source_folder, dest_folder, source_mesh, dest_mesh, debug=False):
    """
    Maps all vtx files from one folder to another folder with two different meshes as a base.
    vtx files in source folder has to be originated from the source mesh.
    All the vtx files in the dest_folder will be in correspondence to the dest_mesh
    :param debug: If true all the vtx files will also be stored as .pts file for better inspection
    :param source_folder:Path to a folder with vtx files
    :param dest_folder: Path to a folder where the results are stored
    :param source_mesh: original vtk mesh
    :param dest_mesh: mesh for the reference of the newly generated points
    :return:
    """
    num_converted_files = 0
    folder = Path(source_folder)
    for file in folder.glob("ids_*.vtx"):
        if file.is_file():  # Ensure it's a file
            print(f"Processing file: {file.name}")
            # Read the content of the .vtx file
            vtx_data = vtx_reader(file)
            remapped_ids = remap_ids(source_mesh, dest_mesh, vtx_data)
            write_to_vtx(f"{dest_folder}/{file.name}", remapped_ids)
            num_converted_files += 1
            if debug:
                write_to_pts(f"{dest_folder}/{file.name}.pts",
                             [dest_mesh.GetPoint(closest_id) for closest_id in remapped_ids]
                             )
    print(f"Number of converted files: {num_converted_files}")


def remap_ids(source_mesh, dest_mesh, source_point_ids):
    """
    Maps source point_id_s coming from the source_mesh to ids from the dest_mesh.
    :param source_mesh: vtk mesh
    :param dest_mesh: vtk mesh
    :param source_point_ids: array
    :return: Array with the new ids
    """
    locator = vtk.vtkPointLocator()
    locator.SetDataSet(dest_mesh)
    locator.BuildLocator()

    new_ids = []
    for source_id in source_point_ids:
        point = source_mesh.GetPoint(source_id)
        closest_id = locator.FindClosestPoint(point)
        new_ids.append(closest_id)

    return new_ids


def ids_to_pts(ids, mesh):
    """
    Converts ids to pts which can be exported to a pts file
    :param ids: The point ids for the mesh
    :param mesh:
    :return:
    """
    points = []
    for id in ids:
        points.append(mesh.GetPoint(id))
    return points


def vtx_to_pts(vtx_filename, pts_filename, mesh_filename):
    ids = vtx_reader(vtx_filename)
    mesh = smart_reader(mesh_filename)
    points = ids_to_pts(ids, mesh)
    write_to_pts(pts_filename, points)



import argparse

import numpy as np
import vtk

from ..openCARP.importing import convert_openCARP_to_vtk
from ..vtk_methods.exporting import vtk_polydata_writer


def get_arguments():
    parser = argparse.ArgumentParser(description="Add data from a .dat file as PointData to a VTK mesh.")
    parser.add_argument("--dat_file", type=str, help="Path to the .dat file containing data.")
    parser.add_argument("--mesh_file", type=str, help="Path to the input VTK mesh file.")
    parser.add_argument("--array_name", type=str, help="Name of the data array in the target output mesh.")
    parser.add_argument("--output_file", type=str, help="Path to save the updated VTK mesh.")
    args = parser.parse_args()
    return args


def add_dat_to_openCARP_mesh(dat_file, openCARP_mesh_file, array_name, output_file):
    """
    Adds data from a .dat file as PointData to a openCARP .pts, .elem mesh.

    :param dat_file: Path to the .dat file containing data to add.
    :param openCARP_mesh_file: Path to the input openCARP file e.g Test-> Test.pts and Test.elem has to exist.
    :param array_name: Name of the array which is represented in the .dat file.
    :param output_file: Path to save the updated VTK mesh.
    """
    mesh = convert_openCARP_to_vtk(openCARP_mesh_file + ".pts", openCARP_mesh_file + ".elem")
    add_dat_to_vtk_mesh(dat_file, mesh, array_name, output_file)


def add_dat_to_vtk_mesh(dat_file, mesh, array_name, output_file, offset=0):
    """
    Adds data from a .dat file as PointData to a VTK mesh.
    :param dat_file: Path to the .dat file containing data to add.
    :param mesh: A VTKMesh in the .vtk format
    :param array_name: Name of the array which is represented in the .dat file.
    :param output_file: Path to save the updated VTK mesh.
    :param offset: Offset which is added to the data array.
    :return:
    """
    try:
        data = np.loadtxt(dat_file)
    except Exception as e:
        raise ValueError(f"Error loading .dat file: {e}")
    if mesh.GetNumberOfPoints() != data.size:
        raise ValueError(f"The number of points in the mesh ({mesh.GetNumberOfPoints()}) "
                         f"does not match the size of the data ({data.size}).")
    vtk_array = vtk.vtkFloatArray()
    vtk_array.SetName(array_name)
    vtk_array.SetNumberOfComponents(1)
    vtk_array.SetNumberOfTuples(data.size)
    for i, value in enumerate(data):
        vtk_array.SetValue(i, value + offset)
    mesh.GetPointData().AddArray(vtk_array)
    mesh.GetPointData().SetActiveScalars(array_name)
    vtk_polydata_writer(output_file, mesh)
    print(f"Updated VTK mesh written to: {output_file}")


if __name__ == "__main__":
    args = get_arguments()
    add_dat_to_openCARP_mesh(args.dat_file, args.mesh_file, args.array_name, args.output_file)



from typing import List


class BaseCommand:
    """Base class for commands.

    This class serves as a base for implementing command-line interface commands.
    """

    def to_cmd(self) -> List:
        """Convert command to list of command line arguments.

        Returns
        -------
        List[str]
            List of command line arguments representing this command.

        Raises
        ------
        NotImplementedError
            If the subclass does not implement this method.
        """
        raise NotImplementedError("to_cmd method must be implemented by subclass")

    def create_from_cmd_list(self, cmd_list: List) -> 'BaseCommand':
        """Create command from list of command line arguments.

        Parameters
        ----------
        cmd_list : List[str]
            List of command line arguments to parse.

        Returns
        -------
        BaseCommand
            New instance of command.

        Raises
        ------
        NotImplementedError
            If the subclass does not implement this method.
        """
        raise NotImplementedError("create_from_cmd_list method must be implemented by subclass")



import os

import vtk


def smart_reader(path):
    """
       Reads in vtk, vtp, vtu, ply and obj files to corresponding mesh classes
       From AugmentA modified to be more robust
    :param path: Path to file with extension
    :return:
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} does not exists")

    extension = str(path).split(".")[-1]

    if extension == "vtk":
        data_checker = vtk.vtkDataSetReader()
        data_checker.SetFileName(str(path))
        data_checker.Update()
        data_type_name = data_checker.GetOutput().GetClassName()

        if data_checker.IsFilePolyData():
            reader = vtk.vtkPolyDataReader()
        elif data_checker.IsFileUnstructuredGrid():
            reader = vtk.vtkUnstructuredGridReader()
        else:
            raise ValueError(f"Type not found or implemented. Was a {data_type_name} file")
        _turn_all_read_on(reader)

    elif extension == "vtp":
        reader = vtk.vtkXMLPolyDataReader()
    elif extension == "vtu":
        reader = vtk.vtkXMLUnstructuredGridReader()
    elif extension == "obj":
        reader = vtk.vtkOBJReader()
    elif extension == "ply":
        reader = vtk.vtkPLYReader()
        _turn_all_read_on(reader)
    else:
        raise ValueError(f"Type not found or implemented. File {path} not readable")

    reader.SetFileName(str(path))
    reader.Update()
    output = reader.GetOutput()

    return output


def _turn_all_read_on(reader):
    reader.ReadAllScalarsOn()
    reader.ReadAllVectorsOn()
    reader.ReadAllTensorsOn()
    reader.ReadAllFieldsOn()
    return reader


def vtx_reader(file_path):
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File {file_path} does not exists")

    data = []
    with open(file_path, 'r') as file:
        lines = file.readlines()
        # Skip the first line (number of elements)
        for line in lines[1:]:
            line = line.strip()
            if line.isdigit():
                data.append(int(line))
    return data



import numpy as np


def normalize_vectors(vectors):
    """
    Normalizes an array of vectors.
    :param vectors: An numpy array of vectors
    :return:
    """
    abs_vectors = np.linalg.norm(vectors, axis=1, keepdims=True)
    abs_vectors = np.where(abs_vectors != 0, abs_vectors, 1)
    vectors = vectors / abs_vectors
    return vectors


def get_normalized_cross_product(center_point, point_a, point_b):
    """
    Returns the normalized cross product between three points.
    :param center_point:
    :param point_a:
    :param point_b:
    :return:
    """

    v1 = point_a - center_point
    v2 = point_b - center_point
    norm = np.cross(v1, v2)

    n = np.linalg.norm(norm, axis=0, keepdims=True)
    return norm / n



global center
import numpy as np
import pyvista as pv
import vtk


def are_normals_outside(mesh):
    """
    Determine if the normals of the mesh point outside the center of the mesh.
    Based on stochastic evaluation of 100 points to be able to process large meshes fast.
    :param mesh: A vtk mesh object
    :return: True if the normals of the mesh point outside, false otherwise
    """
    mesh_with_normals = generate_normals(mesh)

    pv_mesh = pv.wrap(mesh_with_normals)
    points = pv_mesh.points
    normals = pv_mesh.point_data['Normals']

    sample_normals, sample_points = get_sampled_points_and_normals(normals, points)

    mesh_center = points.mean(axis=0)
    vectors_center_to_points = sample_points - mesh_center
    dot_products = np.sum(sample_normals * vectors_center_to_points, axis=1)
    if np.mean(dot_products) < 0:
        print("Normals point inward.")
        return False
    else:
        print("Normals point outward.")
        return True


def get_sampled_points_and_normals(normals, points, number_of_samples=1000):
    """
    Get a random subset of points and normals.
    :param normals: Array of all normals as n-dimensional array
    :param points: Array of all points as n-dimensional array
    :param number_of_samples: How many points/normals should be sampled
    :return:
    """
    num_points = points.shape[0]
    sample_indices = np.random.choice(num_points, min(number_of_samples, num_points), replace=False)
    sample_points = points[sample_indices]
    sample_normals = normals[sample_indices]
    return sample_normals, sample_points


def generate_normals(mesh):
    """
    Generates the normals for a given vtk polydata mesh
    :param mesh: Polydata mesh with normals as array
    :return:
    """
    normal_generator = vtk.vtkPolyDataNormals()
    normal_generator.SetInputData(mesh)
    normal_generator.ComputePointNormalsOn()
    normal_generator.ComputeCellNormalsOff()
    normal_generator.Update()
    polydata_with_normals = normal_generator.GetOutput()
    return polydata_with_normals



import vtk

from ..vtk_methods.converters import vtk_to_numpy


def find_closest_point(dataset, point):
    loc = vtk.vtkPointLocator()
    loc.SetDataSet(dataset)
    loc.BuildLocator()
    return loc.FindClosestPoint(point)


def get_global_cell_ids(mesh, id_array_name="Global_ids"):
    return get_cell_ids(mesh, id_array_name).astype(int)


def get_cell_ids(mesh, id_array_name):
    return vtk_to_numpy(mesh.GetCellData().GetArray(id_array_name))


def get_points(mesh):
    return vtk_to_numpy(mesh.GetPoints().GetData())



import os
import pytest

from src.vtk_openCARP_methods_ibt.openCARP.exporting import write_mesh_from_vtk_obj
from src.vtk_openCARP_methods_ibt.vtk_methods.exporting import vtk_polydata_writer
from src.vtk_openCARP_methods_ibt.openCARP.importing import convert_openCARP_to_vtk, convert_openCARP_to_vtk_single_name
from tests.openCARP.test_writing import _setup_basic_geom, _setup_fibers, _setup_sheet, _setup_elm_tag

SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))

DATA_DIR = SCRIPT_PATH + "/../../data"


def test_working_conversion():
    test_file = f"{DATA_DIR}/cube"
    resulting_mesh = convert_openCARP_to_vtk(test_file + ".pts", test_file + ".elem", test_file + ".lon")
    vtk_polydata_writer(test_file + ".vtk", resulting_mesh)


def test_wrong_filename():
    test_file = f"{DATA_DIR}/cube_wrong"
    with pytest.raises(FileNotFoundError):
        convert_openCARP_to_vtk(test_file + ".pts", test_file + ".elem", test_file + ".lon")


def test_direct_openCARP_importing():
    grid = _setup_geom()
    read_grid = convert_openCARP_to_vtk_single_name(f"{DATA_DIR}/test_opencarp")
    _assert_vtk_objects_equal(grid, read_grid)


def test_direct_openCARP_importing_no_sheet():
    grid = _setup_geom(f"{DATA_DIR}/test_opencarp_nosheet")
    read_grid = convert_openCARP_to_vtk_single_name(f"{DATA_DIR}/test_opencarp_nosheet")
    _assert_vtk_objects_equal(grid, read_grid)


def _setup_geom(file_name: str = f"{DATA_DIR}/test_opencarp"):
    grid = _setup_basic_geom()
    _setup_fibers(grid)
    _setup_sheet(grid)
    _setup_elm_tag(grid)

    # Export to openCARP format
    write_mesh_from_vtk_obj(file_name, grid)
    return grid


def _assert_vtk_objects_equal(obj1, obj2):
    """Compare two VTK objects for equality by checking points and cells."""
    assert obj1.GetNumberOfPoints() == obj2.GetNumberOfPoints(), "Different number of points"
    assert obj1.GetNumberOfCells() == obj2.GetNumberOfCells(), "Different number of cells"

    # Compare points
    for i in range(obj1.GetNumberOfPoints()):
        assert obj1.GetPoint(i) == obj2.GetPoint(i), f"Point {i} differs"

    # Compare cells
    for i in range(obj1.GetNumberOfCells()):
        cell1 = obj1.GetCell(i)
        cell2 = obj2.GetCell(i)
        assert cell1.GetNumberOfPoints() == cell2.GetNumberOfPoints(), f"Cell {i} has different number of points"
        for j in range(cell1.GetNumberOfPoints()):
            assert cell1.GetPointId(j) == cell2.GetPointId(j), f"Cell {i} point {j} differs"



import numpy as np
import vtk

from .converters import vtk_to_numpy, numpy_to_vtk
from ..vtk_methods.filters import get_vtk_geom_filter_port, clean_polydata, get_cells_with_ids


def get_maximum_distance_of_points(points, point_center):
    """
    Returns the maximum distance to the point_center.
    :param points: Has to be a vtk object on which GetPoints() can be executed
    :param point_center:
    :return: the maximum distance between the point_center and the points
    """
    valve_pts = vtk_to_numpy(points.GetPoints().GetData())
    max_dist = 0
    for l in range(len(valve_pts)):
        if np.sqrt(np.sum((point_center - valve_pts[l]) ** 2, axis=0)) > max_dist:
            max_dist = np.sqrt(np.sum((point_center - valve_pts[l]) ** 2, axis=0))

    return max_dist


def cut_mesh_with_radius(mesh, valve_center, max_cutting_radius):
    el_to_del_tot = find_elements_within_radius(mesh, valve_center, max_cutting_radius)
    return cut_elements_from_mesh(mesh, el_to_del_tot)


def cut_elements_from_mesh(mesh, elem_to_delete):
    cell_id_all = list(range(mesh.GetNumberOfCells()))
    el_diff = list(set(cell_id_all).difference(elem_to_delete))

    geo_filter_port, geo_filter = get_vtk_geom_filter_port(get_cells_with_ids(mesh, el_diff))

    return clean_polydata(geo_filter_port, input_is_connection=True)


def find_elements_within_radius(mesh, points_data, radius):
    mesh_id_list = find_points_within_radius(mesh, points_data, radius)

    mesh_cell_id_list = vtk.vtkIdList()
    mesh_cell_temp_id_list = vtk.vtkIdList()
    for i in range(mesh_id_list.GetNumberOfIds()):
        mesh.GetPointCells(mesh_id_list.GetId(i), mesh_cell_temp_id_list)
        for j in range(mesh_cell_temp_id_list.GetNumberOfIds()):
            mesh_cell_id_list.InsertNextId(mesh_cell_temp_id_list.GetId(j))

    id_set = set()
    for i in range(mesh_cell_id_list.GetNumberOfIds()):
        id_set.add(mesh_cell_id_list.GetId(i))

    return id_set


def find_points_within_radius(mesh, center_point, radius):
    """
    Gets points within a radius around the center_point.
    :param mesh: Has to be a vtk mesh
    :param center_point: point in coordinates (x,y,z)
    :param radius: Radius in units of the mesh
    :return:
    """
    locator = vtk.vtkStaticPointLocator()
    locator.SetDataSet(mesh)
    locator.BuildLocator()
    mesh_id_list = vtk.vtkIdList()
    locator.FindPointsWithinRadius(radius, center_point, mesh_id_list)
    return mesh_id_list


def add_vectors_to_vtk(mesh, vectors, name, is_cell_array=True):
    """
    Adds a set of vectors to a vtk object as a cell array with a given name.
    :param mesh: A vtk object.
    :param vectors: A numpy based array of vectors.
    :param name: The name of the resulting cell array
    :param is_cell_array: If True, the resulting array is added to cell data otherwise to point data.
    :return:
    """
    vtk_array = numpy_to_vtk(vectors, deep=True)
    vtk_array.SetName(name)
    if is_cell_array:
        mesh.GetCellData().AddArray(vtk_array)
    else:
        mesh.GetPointData().AddArray(vtk_array)



import vtk

from ..vtk_methods import thresholding
from ..vtk_methods.filters import apply_vtk_geom_filter, clean_polydata
from ..vtk_methods.init_objects import init_connectivity_filter, ExtractionModes

vtk_version = vtk.vtkVersion.GetVTKSourceVersion().split()[-1].split('.')[0]




def extract_largest_region(mesh):
    connect = init_connectivity_filter(mesh, ExtractionModes.LARGEST_REGION)
    surface = apply_vtk_geom_filter(connect.GetOutput())

    return clean_polydata(surface)


from enum import IntEnum


class ThresholdMode(IntEnum):
    UPPER = 0
    LOWER = 1
    BETWEEN = 2


def vtk_thr(model, mode: ThresholdMode | int, points_cells, array, thr1, thr2="None"):
    """
    DEPRECATED use vtk_opencarp_helper_methods.vtk.thresholding instead
    
    Args:
        mode: ThresholdMode enum or int (0=UPPER, 1=LOWER, 2=BETWEEN)
    """
    if mode == ThresholdMode.UPPER:
        thresh = thresholding.get_upper_threshold(model, thr1, "vtkDataObject::FIELD_ASSOCIATION_" + points_cells,
                                                  array)

    elif mode == ThresholdMode.LOWER:
        thresh = thresholding.get_lower_threshold(model, thr1, "vtkDataObject::FIELD_ASSOCIATION_" + points_cells,
                                                  array)

    elif mode == ThresholdMode.BETWEEN:
        thresh = thresholding.get_threshold_between(model, thr1, thr2,
                                                    "vtkDataObject::FIELD_ASSOCIATION_" + points_cells,
                                                    array)
    else:
        raise ValueError("Invalid mode")
    output = thresh.GetOutput()

    return output



import vtk
from pyvista.core import grid

from src.vtk_openCARP_methods_ibt.openCARP.exporting import write_mesh_from_vtk_obj


def test_vtk_to_openCARP():
    """
    Test creating a VTK file and exporting to openCARP format
    """
    grid = _setup_basic_geom()
    _setup_fibers(grid)
    _setup_sheet(grid)
    _setup_elm_tag(grid)

    # Export to openCARP format
    write_mesh_from_vtk_obj("test_opencarp",grid)
    # Verify exported files exist and contain correct data
    with open("test_opencarp.pts", "r") as f:
        pts_content = f.readlines()
    assert len(pts_content) > 0

    with open("test_opencarp.lon", "r") as f:
        lon_content = f.readlines()
    assert len(lon_content) > 0
    assert lon_content[0] == "2\n"
    assert "1.0000 0.0000 0.0000" in lon_content[1]

    with open("test_opencarp.elem", "r") as f:
        elem_content = f.readlines()
    assert len(elem_content) > 0
    assert "Tr" in elem_content[1]

def test_without_sheet():
    grid = _setup_basic_geom()
    _setup_fibers(grid)
    _setup_elm_tag(grid)

    write_mesh_from_vtk_obj("test_opencarp", grid)

    with open("test_opencarp.lon", "r") as f:
        lon_content = f.readlines()
    assert lon_content[0] == "1\n"
    assert len(lon_content) > 0
    assert "1.0000 0.0000 0.0000" in lon_content[1]




def _setup_elm_tag(grid):
    # Add element tags
    elemTag = vtk.vtkIntArray()
    elemTag.SetName("elemTag")
    elemTag.SetNumberOfComponents(1)
    elemTag.SetNumberOfTuples(grid.GetNumberOfCells())
    for i in range(grid.GetNumberOfCells()):
        elemTag.SetValue(i, 1)
    grid.GetCellData().AddArray(elemTag)


def _setup_sheet(grid):
    sheet = vtk.vtkDoubleArray()
    sheet.SetNumberOfComponents(3)
    sheet.SetName("sheet")
    for i in range(grid.GetNumberOfCells()):
        sheet.InsertNextTuple3(0.0, 1.0, 0.0)
    grid.GetCellData().AddArray(sheet)


def _setup_fibers(grid):
    fibers = vtk.vtkDoubleArray()
    fibers.SetNumberOfComponents(3)
    fibers.SetName("fiber")
    for i in range(grid.GetNumberOfCells()):
        fibers.InsertNextTuple3(1.0, 0.0, 0.0)
    grid.GetCellData().AddArray(fibers)


def _setup_basic_geom():
    # Create VTK points
    points = vtk.vtkPoints()
    points.InsertNextPoint(0.0, 0.0, 0.0)
    points.InsertNextPoint(1.0, 0.0, 0.0)
    points.InsertNextPoint(0.0, 1.0, 0.0)
    # Create polydata
    polydata = vtk.vtkPolyData()
    polydata.SetPoints(points)
    # Create triangle
    triangle = vtk.vtkTriangle()
    triangle.GetPointIds().SetId(0, 0)
    triangle.GetPointIds().SetId(1, 1)
    triangle.GetPointIds().SetId(2, 2)
    # Add cell to grid
    grid = vtk.vtkUnstructuredGrid()
    grid.SetPoints(points)
    grid.InsertNextCell(triangle.GetCellType(), triangle.GetPointIds())
    return grid



import re
from .abstract_command import BaseCommand

class LATCommand(BaseCommand):
    def __init__(self, threshold, id_name="ACTs", positional_index=0, all_flag=0, measurand=0, mode=0):
        self.positional_index = positional_index
        self.id_name = id_name
        self.all = all_flag
        self.measurand = measurand
        self.mode = mode
        self.threshold = threshold

    def to_cmd(self):
        """Convert LAT object to list format."""
        return [
            '-num_LATs', 1,
            f'-lats[{self.positional_index}].ID', self.id_name,
            f'-lats[{self.positional_index}].all', self.all,
            f'-lats[{self.positional_index}].measurand', self.measurand,
            f'-lats[{self.positional_index}].mode', self.mode,
            f'-lats[{self.positional_index}].threshold', self.threshold
        ]

    @classmethod
    def create_from_cmd_list(cls, lat_list):
        """Create LAT object from list format."""
        lat_dict = {}
        for i in range(0, len(lat_list), 2):
            key = lat_list[i]
            value = lat_list[i + 1]
            match = re.search(r'\[(\d+)\]', key)
            if '.ID' in key:
                lat_dict['id_name'] = value
            elif '.all' in key:
                lat_dict['all_flag'] = value
            elif '.measurand' in key:
                lat_dict['measurand'] = value
            elif '.mode' in key:
                lat_dict['mode'] = value
            elif '.threshold' in key:
                lat_dict['threshold'] = value
            if match:
                lat_dict['positional_index'] = int(match.group(1))

        return cls(**lat_dict)

    def __str__(self):
        """String representation of LAT object."""
        return (f"LAT(id='{self.id_name}', all={self.all}, "
                f"measurand={self.measurand}, mode={self.mode}, "
                f"threshold={self.threshold})")



import argparse
import os

import numpy as np
import vtk

from ..vtk_methods.helper_methods import add_vectors_to_vtk
from ..vtk_methods.exporting import vtk_polydata_writer


def load_pts(pts_file):
    """
    Loads a .pts file and converts it to an vtkPoints array
    The file is expected to start with the number of points, then lists the point coordinates.
    :param pts_file: the .pts file name with extension .pts
    :return: vtkPoints array
    """
    if not os.path.isfile(pts_file):
        raise FileNotFoundError(f"The .pts file {pts_file} does not exist!")
    if not pts_file.endswith(".pts"):
        raise ValueError(f"The .pts file {pts_file} is not a .pts file!")

    vtk_points = vtk.vtkPoints()
    pts_data = np.loadtxt(pts_file, skiprows=1, usecols=(0, 1, 2))
    num_points = int(np.loadtxt(pts_file, max_rows=1))

    if pts_data.shape[0] != num_points:
        raise ValueError("The number of points declared is not equal to the number of points in the .pts file!")

    for point in pts_data:
        vtk_points.InsertNextPoint(point)
    return vtk_points


def load_elem(elem_file_path):
    """
    Converts a .elem file into a vtkCellArray which is returned
    :param elem_file_path: Element file path with .elem extension
    :return:
    """
    if not os.path.isfile(elem_file_path):
        raise FileNotFoundError(f"The .elem file {elem_file_path} does not exist!")
    if not elem_file_path.endswith(".elem"):
        raise ValueError(f"The .elem file {elem_file_path} is not a .elem file!")

    with open(elem_file_path, 'r') as f:
        lines = f.readlines()

    num_elements = int(lines[0])
    vtk_cells = vtk.vtkCellArray()

    element_tags = []
    for line in lines[1:]:
        parts = line.split()
        num_pts_per_cell = len(parts) - 2

        cell = vtk.vtkIdList()
        for i in range(1, num_pts_per_cell + 1):
            cell.InsertNextId(int(parts[i]))
        vtk_cells.InsertNextCell(cell)
        element_tags.append(int(parts[-1]))

    if vtk_cells.GetNumberOfCells() != num_elements:
        raise ValueError("The number of elements declared is not equal to the number of elements in the .elem file!")
    return vtk_cells, element_tags


def load_lon(lon_file_path):
    if not os.path.exists(lon_file_path):
        raise FileNotFoundError(f"The file {lon_file_path} does not exist")
    with open(lon_file_path, "r") as file:
        lines = file.readlines()

    num_arrays = int(lines[0].strip())
    data = np.loadtxt(lines[1:])

    if num_arrays == 1:
        fibers = data.reshape(-1, 3)
        sheets = None
    elif num_arrays == 2:
        fibers = data[:3]
        sheets = data[3:]
    else:
        raise ValueError("Invalid .lon file: must have 1 or 2 vector arrays.")

    return fibers, sheets

def convert_openCARP_to_vtk_single_name(file_name:str):
    if '.' in file_name:
        VALID_EXTENSIONS = {'pts', 'elem', 'lon',''}

        # Split filename and extension
        base_name, ext = os.path.splitext(file_name)
        ext = ext.lower().lstrip('.')  # remove leading dot and lowercase

        if ext not in VALID_EXTENSIONS:
            raise ValueError(
                f"Invalid file extension: .{ext}. Must be one of: {', '.join(VALID_EXTENSIONS)}"
            )
    else:
        base_name = file_name

    pts_file = f"{base_name}.pts"
    elem_file = f"{base_name}.elem"
    lon_file = f"{base_name}.lon"

    if os.path.exists(lon_file):
        return convert_openCARP_to_vtk(pts_file, elem_file, lon_file)
    else:
        return convert_openCARP_to_vtk(pts_file, elem_file)


def convert_openCARP_to_vtk(pts_file_path, elem_file_path, lon_file_path=""):
    """
    Convert a .pts and .elem file to a .vtk file.
    :param pts_file_path: Filepath of the .pts file with .pts extension
    :param elem_file_path: Filepath of the .elem file with .elem extension
    :param lon_file_path: Filepath of the .lon file with .lon extension. Optional.
    :return: vtk polydata object
    """
    vtk_points = load_pts(pts_file_path)
    vtk_cells, element_tags = load_elem(elem_file_path)

    mesh = vtk.vtkPolyData()
    mesh.SetPoints(vtk_points)
    mesh.SetPolys(vtk_cells)

    cell_data_array = vtk.vtkIntArray()
    cell_data_array.SetName("elemTag")

    for cell_data in element_tags:
        cell_data_array.InsertNextValue(cell_data)

    mesh.GetCellData().SetScalars(cell_data_array)
    if lon_file_path:
        fibers, sheet = load_lon(lon_file_path)
        add_vectors_to_vtk(mesh, fibers, "fiber")
        if sheet is not None:
            add_vectors_to_vtk(mesh, sheet, "sheet")
    return mesh


def parse_args():
    parser = argparse.ArgumentParser(description="Convert openCARP mesh files to VTK format.")

    # Arguments with -- for flexible naming
    parser.add_argument('--pts-file', type=str, required=True, default="data/cube.pte",
                        help="Path to the .pts file containing node positions")
    parser.add_argument('--elem-file', type=str, required=True, default="data/cube.elem",
                        help="Path to the .elem file containing element connectivity")
    parser.add_argument('--lon-file', type=str, required=True, default="data/cube.lon",
                        help="Path to the .lon file containing fiber (and sheet direction)")
    parser.add_argument('--output-file', type=str, required=True, help="Path for the output VTK file")

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()

    vtk_mesh = convert_openCARP_to_vtk(args.pts_file, args.elem_file)
    vtk_polydata_writer(args.output_file, vtk_mesh)



from ..vtk_methods.exporting import vtk_unstructured_grid_writer
#sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from ..vtk_methods.mapper import cell_array_mapper
from ..vtk_methods.reader import smart_reader

rescale_mesh=smart_reader("/Users/pm930l/Projects/data/Joni_volume/LA_vol_with_fiber_res.vtk")
orig_mesh=smart_reader("/Users/pm930l/Projects/data/Joni_volume/LA_vol_with_fiber.vtk")
mapped_mesh=cell_array_mapper(orig_mesh,rescale_mesh,"LA_vol_with_fiber_res_mapped.vtk","all")
vtk_unstructured_grid_writer("/Users/pm930l/Projects/data/Joni_volume/LA_vol_with_fiber_res_mapped.vtk",mapped_mesh)


import vtk
from enum import Enum

from ..mathematical_operations.vector_operations import get_normalized_cross_product


def initialize_plane(norm_1, center):
    plane = vtk.vtkPlane()
    plane.SetNormal(norm_1[0], norm_1[1], norm_1[2])
    plane.SetOrigin(center[0], center[1], center[2])
    return plane


def initialize_plane_with_points(norm_center_point, point_a, point_b, plane_origin, invert_norm=False):
    """
    Creates a plane for given vectors.
    Plane is defined by a perpendicular vector n and a center point (plane_origin)
    The perpendicular vector is the cross product between the norm_center_point and point_a and point_b.
    :param norm_center_point: Center for building the perpendicular vector
    :param point_a:
    :param point_b:
    :param plane_origin:
    :param invert_norm: Inverts the perpendicular vector of the plane
    :return:
    """

    norm_1 = get_normalized_cross_product(norm_center_point, point_a, point_b)
    if invert_norm:
        norm_1 = -norm_1

    return initialize_plane(norm_1, plane_origin)


class ExtractionModes(Enum):
    ALL_REGIONS = "all_regions"
    SPECIFIED_REGIONS = "specified_regions"
    LARGEST_REGION = "largest_region"
    CLOSEST_POINT = "closest_point"
    CLOSEST_POINT_REGION = "closest_point_region"


def init_connectivity_filter(input, extraction_mode: ExtractionModes, color_regions=False, closest_point=None):
    """
    Returns an initialized connectivity filter with update applied
    :param closest_point: Set closest point if the extraction mode is set to the closest point
    :param input: The inout data
    :param extraction_mode: One of the implemented connectivity modes. See class ConnectivityModes
    :param color_regions:
    :return:
    """
    connect = vtk.vtkConnectivityFilter()
    connect.SetInputData(input)
    if extraction_mode == ExtractionModes.ALL_REGIONS:
        connect.SetExtractionModeToAllRegions()
    elif extraction_mode == ExtractionModes.SPECIFIED_REGIONS:
        connect.SetExtractionModeToSpecifiedRegions()
    elif extraction_mode == ExtractionModes.LARGEST_REGION:
        connect.SetExtractionModeToLargestRegion()
    elif extraction_mode == ExtractionModes.CLOSEST_POINT:
        connect.SetExtractionModeToClosestPoint()
    elif extraction_mode == ExtractionModes.CLOSEST_POINT_REGION:
        connect.SetExtractionModeToClosestPointRegion()
    else:
        ValueError("Invalid extraction mode")

    if extraction_mode==ExtractionModes.CLOSEST_POINT or extraction_mode==ExtractionModes.CLOSEST_POINT_REGION:
        assert closest_point.any() is not None, "Closest point not set"
        connect.SetClosestPoint(closest_point)

    if color_regions:
        connect.ColorRegionsOn()
    connect.Update()
    return connect



from vtk import vtkThreshold


def get_threshold_between(input_source, lower_bound, upper_bound, field_association, field_name, idx=0, port=0,
                          connection=0, source_is_input_connection=0, invert=False):
    """

    :param input_source: Input data to threshold. Can be either an input connection or input data format.
    Set source_is_input_connection to 1 if it's an input connection. Automatically swaps variables if lower_bound > upper_bound.
    :param lower_bound:
    :param upper_bound:
    :param field_association: Type of field
    :param field_name:
    :param idx: further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param port:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param connection:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param source_is_input_connection: 1 if it's an input connection, 0 if it's an input data format
    :param invert: True if selection should be inverted
    :return: initialized and updated threshold object
    :return:
    """
    if lower_bound > upper_bound:
        # switch variables
        lower_bound, upper_bound = upper_bound, lower_bound

    thresh = vtkThreshold()
    thresh.SetThresholdFunction(vtkThreshold.THRESHOLD_BETWEEN)

    if source_is_input_connection:
        thresh.SetInputConnection(input_source)
    else:
        thresh.SetInputData(input_source)
    thresh.SetLowerThreshold(lower_bound)
    thresh.SetUpperThreshold(upper_bound)
    if invert:
        thresh.InvertOn()
    thresh.SetInputArrayToProcess(idx, port, connection, field_association, field_name)
    thresh.Update()
    return thresh


def get_lower_threshold(input_source, lower_bound, field_association, field_name, idx=0, port=0,
                        connection=0, source_is_input_connection=0):
    """
    Encapsulates vtk lower threshold

    :param input_source: Input data to threshold. Can be either an input connection or input data format.
    Set source_is_input_connection to 1 if it's an input connection
    :param lower_bound: lower bound for filtering the input source
    :param field_association: Type of field
    :param field_name:
    :param idx: further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param port:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param connection:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param source_is_input_connection: 1 if it's an input connection, 0 if it's an input data format
    :return: initialized and updated threshold object
    """

    thresh = vtkThreshold()

    if source_is_input_connection:
        thresh.SetInputConnection(input_source)
    else:
        thresh.SetInputData(input_source)

    return _standard_thresholding(thresh, vtkThreshold.THRESHOLD_LOWER, lower_bound, field_association, field_name,
                                  idx, port, connection)


def get_upper_threshold(input_source, upper_bound, field_association, field_name, idx=0, port=0, connection=0,
                        source_is_input_connection=0):
    """
    Encapsulates vtk lower threshold

    :param input_source: Input data to threshold. Can be either an input connection or input data format.
    Set source_is_input_connection to 1 if it's an input connection
    :param upper_bound: lower bound for filtering the input source
    :param field_association: Type of field
    :param field_name:
    :param idx: further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param port:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param connection:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param source_is_input_connection: 1 if it's an input connection, 0 if it's an input data format
    :return: initialized and updated threshold object
    """

    thresh = vtkThreshold()
    if source_is_input_connection:
        thresh.SetInputConnection(input_source)
    else:
        thresh.SetInputData(input_source)

    return _standard_thresholding(thresh, vtkThreshold.THRESHOLD_UPPER, upper_bound, field_association, field_name,
                                  idx, port, connection)


def _standard_thresholding(thresh, thresh_function, bound_value, field_association, field_name, idx=0, port=0,
                           connection=0):
    """
    Thresholding generator for lower and upper threshold.
    Should not be used outside these script, for thresholding use get_upper/lower_threshold functions
    :param thresh: Initialized thresh object
    :param thresh_function: vtkThreshold.THRESHOLD_LOWER or vtkThreshold.THRESHOLD_UPPER
    :param bound_value: upper or lower bound value
    :param field_association: Type of field
    :param field_name:
    :param idx: further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param port:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :param connection:further details see vtk.vtkThreshold.SetInputArrayToProcess()
    :return: Initialized and updated threshold object
    """
    thresh.SetThresholdFunction(thresh_function)
    if thresh_function is vtkThreshold.THRESHOLD_LOWER:
        thresh.SetLowerThreshold(bound_value)
    elif thresh_function is vtkThreshold.THRESHOLD_UPPER:
        thresh.SetUpperThreshold(bound_value)
    else:
        raise ValueError(f"Threshold function {thresh_function} is not supported ")

    thresh.SetInputArrayToProcess(idx, port, connection, field_association, field_name)

    thresh.Update()
    return thresh



import vtk

vtk_version = vtk.vtkVersion.GetVTKSourceVersion().split()[-1].split('.')[0]


def apply_vtk_geom_filter(geom, input_is_connection=False):
    geo_filter = vtk.vtkGeometryFilter()
    if input_is_connection:
        geo_filter.SetInputConnection(geom)
    else:
        geo_filter.SetInputData(geom)
    geo_filter.Update()
    return geo_filter.GetOutput()


def get_vtk_geom_filter_port(geom, input_is_connection=False):
    """
    Initializes an vtkGeometryFilter and returns the filter and the output port.
    Filter has to be returned otherwise it will be destroyed by the garbage collector
    :param geom: Input data or connection
    :param input_is_connection: True is the input is also an output port
    :return: the output port of the geometry filter and the geometry filter itself
    """
    geo_filter = vtk.vtkGeometryFilter()
    if input_is_connection:
        geo_filter.SetInputConnection(geom)
    else:
        geo_filter.SetInputData(geom)
    geo_filter.Update()
    return geo_filter.GetOutputPort(), geo_filter


def get_equidistant_points(mesh, tolerance=1.0, point_merging_on=True, tolerance_on=True):
    """
    Returns equidistant points with elements for given interpoint distance
    :param tolerance_on: Set to true, if a specific tolerance is to be applied
    :param mesh: Mesh on which the filter is to be applied
    :param tolerance: Tolerance given in mm
    :param point_merging_on: Enable point merging of vtk clean filter
    :return:
    """
    poly_mesh = apply_vtk_geom_filter(mesh)
    clean = vtk.vtkCleanPolyData()
    clean.SetInputData(poly_mesh)
    if tolerance_on:
        clean.SetAbsoluteTolerance(tolerance)
        clean.ToleranceIsAbsoluteOn()
    if point_merging_on:
        clean.PointMergingOn()
    clean.Update()
    return clean.GetOutput()


def clean_polydata(input, input_is_connection=False):
    """
    Encapsulates the vtkCleanPolyData function
    :param input: Either the input mesh or a vtk input connection port
    :param input_is_connection: True if the input is an output port
    :return: applied clean polydata filter
    """
    cln = vtk.vtkCleanPolyData()
    if input_is_connection:
        cln.SetInputConnection(input)
    else:
        cln.SetInputData(input)
    cln.Update()
    return cln.GetOutput()


def vtk_append(geometries, merge_points=False):
    """
    Appends the geometries to one mesh
    :param geometries: List of at least 1 vtk meshes. Supports vtkPolyData and vtkUnstructuredGrid.
    Hast to be the same type for all the geometries. Has to be a list.
    :param merge_points: Set merge points flag of the append filter on or off.
    :return:
    """
    if len(geometries) < 1:
        raise ValueError("Append filter needs at least 1 input meshe")
    append_filter = vtk.vtkAppendFilter()

    for geom in geometries:
        append_filter.AddInputData(geom)

    append_filter.SetMergePoints(merge_points)
    append_filter.Update()
    return append_filter.GetOutput()


def generate_ids(mesh, point_ids_name, cell_ids_name, field_data_on=False):
    cell_id = vtk.vtkIdFilter()
    cell_id.CellIdsOn()
    cell_id.SetInputData(mesh)
    cell_id.PointIdsOn()
    if field_data_on:
        cell_id.FieldDataOn()
    if int(vtk_version) >= 9:
        cell_id.SetPointIdsArrayName(point_ids_name)
        cell_id.SetCellIdsArrayName(cell_ids_name)
    else:
        cell_id.SetIdsArrayName(point_ids_name)

    cell_id.Update()

    return cell_id.GetOutput()


def get_cells_with_ids(mesh, ids):
    id_list = vtk.vtkIdList()
    for var in ids:
        id_list.InsertNextId(var)

    return apply_extract_cell_filter(mesh, id_list)


def apply_extract_cell_filter(input, id_list, input_is_connection=False):
    """
    Extract cells from mesh based on the id_list
    :param input_is_connection: Set to true if the input is an output port
    :param input: Input mesh or connection
    :param id_list:
    :return:
    """
    extract = vtk.vtkExtractCells()
    if input_is_connection:
        extract.SetInputConnection(input)
    else:
        extract.SetInputData(input)
    extract.SetCellList(id_list)
    extract.Update()

    return extract.GetOutput()


def get_center_of_mass(mesh, set_use_scalars_as_weights):
    center_of_mass_filter = vtk.vtkCenterOfMass()
    center_of_mass_filter.SetInputData(mesh)
    center_of_mass_filter.SetUseScalarsAsWeights(set_use_scalars_as_weights)
    center_of_mass_filter.Update()
    return center_of_mass_filter.GetCenter()


def get_feature_edges(input_data, boundary_edges_on, feature_edges_on, manifold_edges_on, non_manifold_edges_on):
    feature_edge_filter = vtk.vtkFeatureEdges()
    feature_edge_filter.SetInputData(input_data)
    if boundary_edges_on:
        feature_edge_filter.BoundaryEdgesOn()
    else:
        feature_edge_filter.BoundaryEdgesOff()
    if feature_edges_on:
        feature_edge_filter.FeatureEdgesOn()
    else:
        feature_edge_filter.FeatureEdgesOff()
    if manifold_edges_on:
        feature_edge_filter.ManifoldEdgesOn()
    else:
        feature_edge_filter.ManifoldEdgesOff()
    if non_manifold_edges_on:
        feature_edge_filter.NonManifoldEdgesOn()
    else:
        feature_edge_filter.NonManifoldEdgesOff()

    feature_edge_filter.Update()
    return feature_edge_filter.GetOutput()


def get_elements_above_plane(input, plane, extract_boundary_cells_on=False, input_is_connection=False):
    """
    Encapsulates the vtkExtractGeometry filter and returns elements which lay above the specified plane.
    :param input_is_connection: Set to true if the input is an output port
    :param input: Input data or connection. Set input_is_connection to true if the input is an output port
    :param plane: vtkPlane
    :param extract_boundary_cells_on: True if the option to extract boundary cells should be turned on
    :return:
    """
    mesh_extract_filter = vtk.vtkExtractGeometry()
    if input_is_connection:
        mesh_extract_filter.SetInputConnection(input)
    else:
        mesh_extract_filter.SetInputData(input)
    mesh_extract_filter.SetImplicitFunction(plane)
    if extract_boundary_cells_on:
        mesh_extract_filter.ExtractBoundaryCellsOn()
    mesh_extract_filter.Update()
    return mesh_extract_filter.GetOutput()


def get_mesh_extract_filter_output(input_data, plane):
    return get_elements_above_plane(input_data, plane)


def translate_poly_data(mesh, translation):
    """
    Translates a polydata based on a three-dimensional translation vector
    :param mesh: VTK polydata mesh
    :param translation: Array with three indices for translation in x,y and z.
    :return:
    """
    if len(translation) != 3:
        raise ValueError(f"Translation must be a three dimensional array but was {len(translation)}")
    transform = vtk.vtkTransform()
    transform.Translate(translation[0], translation[1], translation[2])

    return apply_transform_filter(mesh, transform)


def apply_transform_filter(mesh, transform):
    transform_filter = vtk.vtkTransformPolyDataFilter()
    transform_filter.SetInputData(mesh)
    transform_filter.SetTransform(transform)
    transform_filter.Update()
    return transform_filter.GetOutput()


def scale_poly_data(mesh, scale_factor):
    """
    Scales a polydata based on a scale factor
    :param mesh: VTK polydata mesh
    :param scale_factor: Factor which scales the mesh

    :return:
    """
    transform = vtk.vtkTransform()
    transform.Scale(scale_factor, scale_factor, scale_factor)
    return apply_transform_filter(mesh, transform)



from vtkmodules.vtkCommonDataModel import vtkDataSet
from ..vtk_methods.converters import vtk_to_numpy
from ..vtk_methods.finder import get_cell_ids


def write_to_elem(filename, mesh, tag):
    """

    :param filename: Filename with .elem extension
    :param mesh: Mesh which should be stored as element file.
    :param tag: Tags which are attached to the mesh
    :return:
    """
    if not filename.endswith('.elem'):
        raise ValueError(f'Filename must end with .elem extension but was {filename}')
    with open(filename, "w") as f:
        f.write(f"{mesh.GetNumberOfCells()}\n")
        for i in range(mesh.GetNumberOfCells()):
            cell = mesh.GetCell(i)
            if cell.GetNumberOfPoints() == 2:
                f.write(
                    f"Ln {cell.GetPointIds().GetId(0)} {cell.GetPointIds().GetId(1)} {tag[i]}\n")
            elif cell.GetNumberOfPoints() == 3:
                f.write("Tr {} {} {} {}\n".format(cell.GetPointIds().GetId(0), cell.GetPointIds().GetId(1),
                                                  cell.GetPointIds().GetId(2), tag[i]))
            elif cell.GetNumberOfPoints() == 4:
                f.write("Tt {} {} {} {} {}\n".format(cell.GetPointIds().GetId(0), cell.GetPointIds().GetId(1),
                                                     cell.GetPointIds().GetId(2), cell.GetPointIds().GetId(3),
                                                     tag[i]))
            else:
                print("strange " + str(cell.GetNumberOfPoints()))


def write_to_lon(filename_lon, elem, sheet=None, precession=4):
    if not filename_lon.endswith('.lon'):
        raise ValueError(f'Filename must end with .lon extension but was {filename_lon}')

    if sheet is None:
        rows = (format_vector_to_str(e, precession) + "\n" for e in elem)
        num_arrays=1
    else:
        rows = (format_vector_to_str(e, precession) + " " + format_vector_to_str(s, precession) + "\n"
                for e, s in zip(elem, sheet))
        num_arrays = 2
    with open(filename_lon, "w") as f:
        f.write(f"{num_arrays}\n")
        f.writelines(rows)


def format_vector_to_str(v, precision):
    return " ".join(f"{x:.{precision}f}" for x in v)


def write_to_pts(filename_pts, pts):
    if not filename_pts.endswith('.pts'):
        raise ValueError(f'Filename must end with .pts extension but was {filename_pts}')
    with open(filename_pts, "w") as f:
        f.write(f"{len(pts)}\n")
        for i in range(len(pts)):
            f.write(f"{pts[i][0]} {pts[i][1]} {pts[i][2]}\n")


def write_mesh(filename, pts, mesh: vtkDataSet, elem_tags, fiber_long=None, fiber_sheet=None):
    write_to_pts(filename + ".pts", pts)
    write_to_elem(filename + ".elem", mesh, elem_tags)
    if fiber_long is not None:
        write_to_lon(filename + ".lon", fiber_long, fiber_sheet)


def write_mesh_from_vtk_obj(filename: str, mesh, tag_name="elemTag", fiber_name="fiber", sheet_name="sheet"):
    pts = vtk_to_numpy(mesh.GetPoints().GetData())
    fibers = get_cell_ids(mesh, fiber_name)
    if mesh.GetCellData().HasArray(sheet_name):
        sheet = get_cell_ids(mesh, sheet_name)
    else:
        sheet = None
    tags = get_cell_ids(mesh, tag_name)
    write_mesh(filename, pts, mesh, tags, fibers, sheet)
